"""Utilities for using sizzle (jQuery-style) selectors with Selenium."""

import json

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.support.ui import WebDriverWait

from .has_driver import exception_indicates_stale_element

SIZZLE_LOAD_TIMEOUT = 5
SIZZLE_URL = "//cdnjs.cloudflare.com/ajax/libs/sizzle/1.10.18/sizzle.js"


def sizzle_selector_clickable(selector):
    def ec(driver):
        elements = find_elements_by_sizzle(driver, selector)
        if not elements:
            return False
        element = elements[0]
        try:
            clickable = element.is_displayed() and element.is_enabled()
        except Exception as e:
            # Handle the case where the element is detached between when it is
            # discovered and when it is checked - it is likely changing quickly
            # and the next pass will be the final state. If not, this should be
            # wrapped in a wait anyway - so no problems there. For other
            # non-custom selectors I believe this all happens on the Selenium
            # server and so there is likely no need to handle this case - they
            # are effectively atomic.
            if exception_indicates_stale_element(e):
                return None

            raise
        if clickable:
            return element
        else:
            return None

    return ec


def sizzle_presence_of_selector(selector):
    def ec(driver):
        elements = find_elements_by_sizzle(driver, selector)
        if not elements:
            return False
        element = elements[0]
        try:
            displayed = element.is_displayed()
        except Exception as e:
            # See note above insizzle_selector_clickable about this exception.
            if exception_indicates_stale_element(e):
                return None

            raise

        if displayed:
            return element
        else:
            return None

    return ec


def find_element_by_sizzle(driver, sizzle_selector: str):
    """
    Finds an element by sizzle selector.

    :param sizzle_selector: The sizzle selector to use when finding element.
    """
    elements = driver.find_elements_by_sizzle(sizzle_selector)
    if elements:
        return elements[0]
    else:
        raise NoSuchElementException(
            f"Unable to locate element by Sizzle: {sizzle_selector}"
        )


def find_elements_by_sizzle(driver, sizzle_selector: str):
    """
    Finds elements by sizzle selector.

    :param sizzle_selector: The sizzle selector to use when finding elements.
    """
    if not _is_sizzle_loaded(driver):
        _inject_sizzle(driver, SIZZLE_URL, SIZZLE_LOAD_TIMEOUT)
    elements = driver.execute_script(_make_sizzle_string(sizzle_selector))
    return elements


def _inject_sizzle(driver, sizzle_url, timeout):
    script = """
        if(typeof(window.$) != "undefined") {
            // Just reuse jQuery if it is available, avoids potential amd problems
            // that have cropped up with Galaxy for instance.
            window.Sizzle = window.$;
        } else {
            var _s = document.createElement("script");
            _s.type = "text/javascript";
            _s.src = "%s";
            var _h = document.getElementsByTagName("head")[0];
            _h.appendChild(_s);
        }
    """ % sizzle_url
    driver.execute_script(script)
    wait = WebDriverWait(driver, timeout)
    wait.until(lambda d: _is_sizzle_loaded(d),
               f"Can't inject Sizzle in {timeout} seconds")


def _is_sizzle_loaded(driver):
    script = "return typeof(Sizzle) != \"undefined\";"
    return driver.execute_script(script)


def _make_sizzle_string(sizzle_selector):
    # Use json.dumps to escape quotes
    selector = json.dumps(sizzle_selector)
    return f"return Sizzle({selector});"


__all__ = (
    "find_element_by_sizzle",
    "find_elements_by_sizzle",
    "sizzle_selector_clickable",
    "sizzle_presence_of_selector",
)



"""A mixin to extend a class that has self.driver with higher-level constructs.

This should be mixed into classes with a self.driver and self.default_timeout
attribute.
"""

from selenium.common.exceptions import TimeoutException as SeleniumTimeoutException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait

UNSPECIFIED_TIMEOUT = object()


class HasDriver:
    driver: WebDriver

    def assert_xpath(self, xpath):
        assert self.driver.find_element_by_xpath(xpath)

    def assert_selector(self, selector):
        assert self.driver.find_element_by_css_selector(selector)

    def assert_selector_absent_or_hidden(self, selector):
        elements = self.driver.find_elements_by_css_selector(selector)
        for element in elements:
            assert not element.is_displayed()

    def assert_absent_or_hidden(self, selector_template):
        elements = self.find_elements(selector_template)
        for element in elements:
            assert not element.is_displayed()

    def selector_is_displayed(self, selector):
        element = self.driver.find_element_by_css_selector(selector)
        return element.is_displayed()

    def is_displayed(self, selector_template):
        element = self.driver.find_element(*selector_template.element_locator)
        return element.is_displayed()

    def assert_selector_absent(self, selector):
        assert len(self.driver.find_elements_by_css_selector(selector)) == 0

    def find_elements(self, selector_template):
        return self.driver.find_elements(*selector_template.element_locator)

    def assert_absent(self, selector_template):
        assert len(self.find_elements(selector_template)) == 0

    def element_absent(self, selector_template):
        return len(self.find_elements(selector_template)) == 0

    def wait_for_xpath(self, xpath, **kwds):
        element = self._wait_on(
            ec.presence_of_element_located((By.XPATH, xpath)),
            f"XPATH selector [{xpath}] to become present",
            **kwds
        )
        return element

    def wait_for_xpath_visible(self, xpath, **kwds):
        element = self._wait_on(
            ec.visibility_of_element_located((By.XPATH, xpath)),
            f"XPATH selector [{xpath}] to become visible",
            **kwds
        )
        return element

    def wait_for_selector(self, selector, **kwds):
        element = self._wait_on(
            ec.presence_of_element_located((By.CSS_SELECTOR, selector)),
            f"CSS selector [{selector}] to become present",
            **kwds
        )
        return element

    def wait_for_present(self, selector_template, **kwds):
        element = self._wait_on(
            ec.presence_of_element_located(selector_template.element_locator),
            f"{selector_template.description} to become present",
            **kwds
        )
        return element

    def wait_for_visible(self, selector_template, **kwds):
        element = self._wait_on(
            ec.visibility_of_element_located(selector_template.element_locator),
            f"{selector_template.description} to become visible",
            **kwds
        )
        return element

    def wait_for_selector_visible(self, selector, **kwds):
        element = self._wait_on(
            ec.visibility_of_element_located((By.CSS_SELECTOR, selector)),
            f"CSS selector [{selector}] to become visible",
            **kwds
        )
        return element

    def wait_for_selector_clickable(self, selector, **kwds):
        element = self._wait_on(
            ec.element_to_be_clickable((By.CSS_SELECTOR, selector)),
            f"CSS selector [{selector}] to become clickable",
            **kwds
        )
        return element

    def wait_for_clickable(self, selector_template, **kwds):
        element = self._wait_on(
            ec.element_to_be_clickable(selector_template.element_locator),
            f"{selector_template.description} to become clickable",
            **kwds
        )
        return element

    def wait_for_selector_absent_or_hidden(self, selector, **kwds):
        element = self._wait_on(
            ec.invisibility_of_element_located((By.CSS_SELECTOR, selector)),
            f"CSS selector [{selector}] to become absent or hidden",
            **kwds
        )
        return element

    def wait_for_selector_absent(self, selector, **kwds):
        element = self._wait_on(
            lambda driver: len(driver.find_elements_by_css_selector(selector)) == 0,
            f"CSS selector [{selector}] to become absent",
            **kwds
        )
        return element

    def wait_for_absent(self, selector_template, **kwds):
        element = self._wait_on(
            lambda driver: len(driver.find_elements(*selector_template.element_locator)) == 0,
            f"{selector_template.description} to become absent",
            **kwds
        )
        return element

    def wait_for_absent_or_hidden(self, selector_template, **kwds):
        element = self._wait_on(
            ec.invisibility_of_element_located(selector_template.element_locator),
            f"{selector_template.description} to become absent or hidden",
            **kwds
        )
        return element

    def wait_for_id(self, id, **kwds):
        return self._wait_on(
            ec.presence_of_element_located((By.ID, id)),
            f"presence of DOM ID [{id}]",
            **kwds
        )

    def click(self, selector_template):
        element = self.driver.find_element(*selector_template.element_locator)
        element.click()

    def _timeout_message(self, on_str):
        return f"Timeout waiting on {on_str}."

    def _wait_on(self, condition, on_str=None, **kwds):
        if on_str is None:
            on_str = str(condition)
        wait = self.wait(**kwds)
        return wait.until(condition, self._timeout_message(on_str))

    def action_chains(self):
        return ActionChains(self.driver)

    def send_enter(self, element):
        element.send_keys(Keys.ENTER)

    def send_escape(self, element):
        element.send_keys(Keys.ESCAPE)

    def send_backspace(self, element):
        element.send_keys(Keys.BACKSPACE)

    def timeout_for(self, **kwds) -> float:
        ...

    def wait(self, timeout=UNSPECIFIED_TIMEOUT, **kwds):
        if timeout is UNSPECIFIED_TIMEOUT:
            timeout = self.timeout_for(**kwds)
        return WebDriverWait(self.driver, timeout)

    def click_xpath(self, xpath):
        element = self.driver.find_element_by_xpath(xpath)
        element.click()

    def click_label(self, text):
        element = self.driver.find_element_by_link_text(text)
        element.click()

    def click_selector(self, selector):
        element = self.driver.find_element_by_css_selector(selector)
        element.click()

    def fill(self, form, info):
        for key, value in info.items():
            input_element = form.find_element_by_name(key)
            input_element.send_keys(value)

    def click_submit(self, form):
        submit_button = form.find_element_by_css_selector("input[type='submit']")
        submit_button.click()

    def prepend_timeout_message(self, timeout_exception, message):
        msg = message
        timeout_msg = timeout_exception.msg
        if timeout_msg:
            msg += f" {timeout_msg}"
        return SeleniumTimeoutException(
            msg=msg,
            screen=timeout_exception.screen,
            stacktrace=timeout_exception.stacktrace,
        )


def exception_indicates_not_clickable(exception):
    return "not clickable" in str(exception)


def exception_indicates_stale_element(exception):
    return "stale" in str(exception)


__all__ = (
    "exception_indicates_not_clickable",
    "exception_indicates_stale_element",
    "HasDriver",
    "SeleniumTimeoutException",
)



def restrict_test(context, section):
    """
    Disable the Test Section section

    This tool filter will disable the Test Section section.
    """
    if section.name == 'Test Section':
        return False
    return True



from .context import GalaxySeleniumContextImpl, init as c_init


def init(config=None):
    return c_init(config=config, clazz=JupyterContextImpl)


class JupyterContextImpl(GalaxySeleniumContextImpl):

    def screenshot(self, label):
        path = super().screenshot(label)
        from IPython.display import Image
        return Image(filename=path)



from .components import (
    Component,
    Target,
)


class SmartComponent:
    """Wrap a Component with driver aware methods.

    Allows smarter selectors that know how to wait for themselves, test themselves,
    click themselves, etc.... More "magic", but much cleaner usage.
    """

    def __init__(self, component, has_driver):
        self._component = component
        self._has_driver = has_driver

    def __getattr__(self, name):
        return self._wrap(getattr(self._component, name))

    def __getitem__(self, name):
        return self._wrap(self._component[name])

    def _wrap(self, simple_object):
        if isinstance(simple_object, Component):
            return SmartComponent(simple_object, self._has_driver)
        elif isinstance(simple_object, Target):
            return SmartTarget(simple_object, self._has_driver)
        else:
            return simple_object


class SmartTarget:
    """Wrap a Target with driver aware methods.
    """

    def __init__(self, target, has_driver):
        self._target = target
        self._has_driver = has_driver

    def __call__(self, *args, **kwds):
        return self._wrap(self._target(*args[:], **kwds.copy()))

    def __getattr__(self, name):
        return self._wrap(getattr(self._target, name))

    def __getitem__(self, name):
        return self._wrap(self._target[name])

    def _wrap(self, simple_object):
        if isinstance(simple_object, Target):
            return SmartTarget(simple_object, self._has_driver)
        else:
            return simple_object

    def all(self):
        return self._has_driver.driver.find_elements(*self._target.element_locator)

    def wait_for_and_click(self, **kwds):
        return self._has_driver.wait_for_and_click(self._target, **kwds)

    def wait_for_visible(self, **kwds):
        return self._has_driver.wait_for_visible(self._target, **kwds)

    def wait_for_clickable(self, **kwds):
        return self._has_driver.wait_for_clickable(self._target, **kwds)

    def wait_for_text(self, **kwds):
        return self._has_driver.wait_for_visible(self._target, **kwds).text

    def wait_for_value(self, **kwds):
        return self._has_driver.wait_for_visible(self._target, **kwds).get_attribute("value")

    @property
    def is_displayed(self):
        return self._has_driver.is_displayed(self._target)

    @property
    def is_absent(self):
        return self._has_driver.element_absent(self._target)

    def wait_for_absent_or_hidden(self, **kwds):
        self._has_driver.wait_for_absent_or_hidden(self._target, **kwds)

    def wait_for_absent(self, **kwds):
        self._has_driver.wait_for_absent(self._target, **kwds)

    def wait_for_present(self, **kwds):
        return self._has_driver.wait_for_present(self._target, **kwds)

    def assert_absent(self, **kwds):
        self._has_driver.assert_absent(self._target, **kwds)

    def assert_absent_or_hidden(self, **kwds):
        self._has_driver.assert_absent_or_hidden(self._target, **kwds)

    def assert_absent_or_hidden_after_transitions(self, **kwds):
        self._has_driver.assert_absent_or_hidden_after_transitions(self._target, **kwds)

    def has_class(self, class_name):
        return class_name in self._has_driver.driver.find_element(*self._target.element_locator).get_attribute("class")

    def wait_for_and_send_keys(self, *text):
        self.wait_for_visible().send_keys(*text)



import re
import string
from abc import (
    ABCMeta,
    abstractproperty,
)
from typing import Union

from selenium.webdriver.common.by import By

from galaxy.util.bunch import Bunch


class Target(metaclass=ABCMeta):

    @abstractproperty
    def description(self):
        """Return a plain-text description of the browser target for logging/messages."""

    @abstractproperty
    def element_locator(self):
        """Return a (by, selector) Selenium elment locator tuple for this selector."""


class SelectorTemplate(Target):

    def __init__(self, selector: str, selector_type: str, children=None, kwds=None, with_classes=None, with_data=None):
        if selector_type == "data-description":
            selector_type = "css"
            selector = f'[data-description="{selector}"]'
        self._selector = selector
        self.selector_type = selector_type
        self._children = children or {}
        self.__kwds = kwds or {}
        self.with_classes = with_classes or []
        self._with_data = with_data or {}

    @staticmethod
    def from_dict(raw_value, children=None):
        children = children or {}
        if isinstance(raw_value, dict):
            return SelectorTemplate(raw_value["selector"], raw_value.get("type", "css"), children=children)
        else:
            return SelectorTemplate(raw_value, "css", children=children)

    def with_class(self, class_):
        assert self.selector_type == "css"
        return SelectorTemplate(self._selector, self.selector_type, kwds=self.__kwds, with_classes=self.with_classes + [class_], with_data=self._with_data.copy(), children=self._children)

    def with_data(self, key, value):
        assert self.selector_type == "css"
        with_data = self._with_data.copy()
        with_data[key] = value
        return SelectorTemplate(self._selector, self.selector_type, kwds=self.__kwds, with_classes=self.with_classes, with_data=with_data, children=self._children)

    def descendant(self, has_selector):
        assert self.selector_type == "css"
        if hasattr(has_selector, "selector"):
            selector = has_selector.selector
        else:
            selector = has_selector

        return SelectorTemplate(f"{self.selector} {selector}", self.selector_type, kwds=self.__kwds, children=self._children)

    def __call__(self, **kwds):
        new_kwds = self.__kwds.copy()
        new_kwds.update(**kwds)
        return SelectorTemplate(self._selector, self.selector_type, kwds=new_kwds, with_classes=self.with_classes, children=self._children)

    @property
    def description(self):
        if self.selector_type == "css":
            template = "CSS selector [%s]"
        elif self.selector_type == "xpath":
            template = "XPATH selector [%s]"
        elif self.selector_type == "id":
            template = "DOM element with id [%s]"
        return template % self.selector

    @property
    def selector(self):
        selector = self._selector
        if self.__kwds is not None:
            selector = string.Template(selector).substitute(self.__kwds)
        selector = selector + "".join(f".{c}" for c in self.with_classes)
        if self._with_data:
            for key, value in self._with_data.items():
                selector = selector + f'[data-{key}="{value}"]'
        return selector

    @property
    def element_locator(self):
        if self.selector_type == "css":
            by = By.CSS_SELECTOR
        elif self.selector_type == "xpath":
            by = By.XPATH
        elif self.selector_type == "id":
            by = By.ID
        else:
            raise Exception(f"Unknown selector type {self.selector_type}")
        return (by, self.selector)

    @property
    def as_css_class(self):
        assert self.selector_type == "css"
        assert re.compile(r"\.\w+").match(self._selector)
        return self._selector[1:]

    def __getattr__(self, name):
        if name in self._children:
            return self._children[name](**{"_": self.selector})
        else:
            raise AttributeError(f"Could not find child [{name}] in {self._children}")

    __getitem__ = __getattr__


class Label(Target):

    def __init__(self, text):
        self.text = text

    @property
    def description(self):
        return f"Link text [{self.text}]"

    @property
    def element_locator(self):
        return (By.LINK_TEXT, self.text)


class Text(Target):

    def __init__(self, text):
        self.text = text

    @property
    def description(self):
        return f"Text containing [{self.text}]"

    @property
    def element_locator(self):
        return (By.PARTIAL_LINK_TEXT, self.text)


HasText = Union[Label, Text]


class Component:

    def __init__(self, name, sub_components, selectors, labels, text):
        self._name = name
        self._sub_components = sub_components
        self._selectors = selectors
        self._labels = labels
        self._text = text

        self.selectors = Bunch(**self._selectors)
        self.labels = Bunch(**self._labels)
        self.text = Bunch(**self._text)

    @property
    def selector(self):
        if "_" in self._selectors:
            return self._selectors["_"]
        else:
            raise Exception(f"No _ selector for [{self}]")

    @staticmethod
    def from_dict(name, raw_value):
        selectors = {}
        labels = {}
        text = {}
        sub_components = {}

        for key, value in raw_value.items():
            if key == "selectors":
                base_selector = None
                if "_" in value:
                    base_selector = value["_"]
                    del value["_"]

                for selector_key, selector_value in value.items():
                    selectors[selector_key] = SelectorTemplate.from_dict(selector_value)

                if base_selector:
                    selectors["_"] = SelectorTemplate.from_dict(base_selector, children=selectors)

            elif key == "labels":
                for label_key, label_value in value.items():
                    labels[label_key] = Label(label_value)
            elif key == "text":
                for text_key, text_value in value.items():
                    text[text_key] = Text(text_value)
            else:
                component = Component.from_dict(key, value)
                sub_components[key] = component

        return Component(name, sub_components, selectors, labels, text)

    def __getattr__(self, attr):
        if attr in self._sub_components:
            return self._sub_components[attr]
        elif attr in self._selectors:
            return self._selectors[attr]
        elif attr in self._labels:
            return self._labels[attr]
        elif attr in self._text:
            return self._text[attr]
        else:
            raise AttributeError(f"Failed to find referenced sub-component/selector/label/text [{attr}]")

    __getitem__ = __getattr__

    def __str__(self):
        return f"Component[{self._name}]"



import logging
import os

try:
    from pyvirtualdisplay import Display
except ImportError:
    Display = None

from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.remote.webdriver import WebDriver

logger = logging.getLogger('selenium.webdriver.remote.remote_connection')
logger.setLevel(logging.WARNING)

DEFAULT_BROWSER = "auto"
DEFAULT_DOWNLOAD_PATH = '/tmp/'
LOGGING_PREFS = {
    "browser": "ALL",
}
DEFAULT_SELENIUM_BROWSER = "auto"
DEFAULT_SELENIUM_REMOTE = False
DEFAULT_SELENIUM_REMOTE_PORT = "4444"
DEFAULT_SELENIUM_REMOTE_HOST = "127.0.0.1"
DEFAULT_WINDOW_WIDTH = 1280
DEFAULT_WINDOW_HEIGHT = 1000
VALID_LOCAL_BROWSERS = ["CHROME", "FIREFOX", "OPERA"]


class ConfiguredDriver:
    driver: WebDriver

    def __init__(
        self,
        browser=DEFAULT_SELENIUM_BROWSER,
        remote=DEFAULT_SELENIUM_REMOTE,
        remote_host=DEFAULT_SELENIUM_REMOTE_HOST,
        remote_port=DEFAULT_SELENIUM_REMOTE_PORT,
        headless=False,
    ):
        self.config = dict(
            browser=browser,
            remote=remote,
            remote_host=remote_host,
            remote_port=remote_port,
            headless=headless,
        )
        if remote:
            driver = get_remote_driver(
                host=remote_host,
                port=remote_port,
                browser=browser,
            )
        else:
            driver = get_local_driver(
                browser=browser,
                headless=headless,
            )
        # Need larger than default window size for workflow index page, paired list creator.
        driver.set_window_size(DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT)
        self.driver = driver

    def to_dict(self):
        return self.config

    @staticmethod
    def from_dict(as_dict):
        return ConfiguredDriver(**as_dict)


def get_local_browser(browser):
    if browser == "auto":
        if _which("chromedriver"):
            browser = "CHROME"
        elif _which("geckodriver"):
            browser = "FIREFOX"
        else:
            raise Exception("Selenium browser is 'auto' but neither geckodriver or chromedriver are found on PATH.")
    return browser


def get_local_driver(browser=DEFAULT_BROWSER, headless=False) -> WebDriver:
    browser = get_local_browser(browser)
    if browser not in VALID_LOCAL_BROWSERS:
        raise AssertionError(f"{browser} not in VALID_LOCAL_BROWSERS ({VALID_LOCAL_BROWSERS})")
    driver_to_class = {
        "CHROME": webdriver.Chrome,
        "FIREFOX": webdriver.Firefox,
        "OPERA": webdriver.Opera,
    }
    driver_class = driver_to_class[browser]
    if browser == 'CHROME':
        options = ChromeOptions()
        if headless:
            options.add_argument('--headless')
        prefs = {'download.default_directory': DEFAULT_DOWNLOAD_PATH}
        options.add_experimental_option('prefs', prefs)
        return driver_class(desired_capabilities={"loggingPrefs": LOGGING_PREFS}, chrome_options=options)
    elif browser == 'FIREFOX':
        fp = webdriver.FirefoxProfile()
        fp.set_preference('network.proxy.type', 2)
        fp.set_preference('network.proxy.autoconfig_url',
                          "http://127.0.0.1:9675")
        fp.set_preference('browser.download.folderList', 2)
        fp.set_preference('browser.download.dir', DEFAULT_DOWNLOAD_PATH)
        fp.set_preference("browser.helperApps.neverAsk.saveToDisk", 'application/octet-stream')
        return driver_class(firefox_profile=fp)

    else:
        return driver_class(desired_capabilities={"loggingPrefs": LOGGING_PREFS})


def get_remote_driver(
    host,
    port,
    browser=DEFAULT_BROWSER
) -> WebDriver:
    # docker run -d -p 4444:4444 -v /dev/shm:/dev/shm selenium/standalone-chrome:3.0.1-aluminum
    if browser == "auto":
        browser = "CHROME"
    assert browser in ["CHROME", "EDGE", "ANDROID", "FIREFOX", "INTERNETEXPLORER", "IPAD", "IPHONE", "OPERA", "SAFARI"]
    desired_capabilities = getattr(DesiredCapabilities, browser)
    desired_capabilities["loggingPrefs"] = LOGGING_PREFS
    executor = f'http://{host}:{port}/wd/hub'
    driver = webdriver.Remote(
        command_executor=executor,
        desired_capabilities=desired_capabilities,
    )
    return driver


def is_virtual_display_available():
    return Display is not None


def virtual_display_if_enabled(enabled):
    if enabled:
        display = Display(visible=0, size=(800, 600))
        display.start()
        return display
    else:
        return NoopDisplay()


class NoopDisplay:

    def stop(self):
        """No-op stop for consistent use with pyvirtualdisplay Display class."""


# Purposely copied from galaxy.util - just use galaxy.util if we decide
# galaxy_selenium should definitely depend on Galaxy/galaxy-lib.
def _which(file):
    # http://stackoverflow.com/questions/5226958/which-equivalent-function-in-python
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(f"{path}/{file}"):
            return f"{path}/{file}"

    return None


__all__ = (
    'ConfiguredDriver',
    'get_local_driver',
    'get_remote_driver',
    'is_virtual_display_available',
    'virtual_display_if_enabled',
)
